/* network.c -- network management stuff
 *
 * author: hackbard@hackdaworld.dyndns.org
 *
 */

#include "network.h"

int network_init(t_net *net,int outfd) {

  dprintf(net->outfd,"[network] initializing network api ...\n");

  memset(net->connection,0,MAX_CONNECTIONS*sizeof(t_connection));
  net->c_count=0;
  net->sendmask=0;
  net->outfd=outfd;

  return N_SUCCESS;
}

int network_listen(t_net *net) {
  
  struct sockaddr_in addr;
  int true;

  if((net->l_fd=socket(AF_INET,SOCK_STREAM,0))==-1) {
    perror("[network] socket call");
    return N_ERROR;
  }

  memset(&addr,0,sizeof(struct sockaddr));
  addr.sin_family=AF_INET;
  addr.sin_port=htons(net->l_port);
  addr.sin_addr.s_addr=INADDR_ANY;

  if(bind(net->l_fd,(struct sockaddr *)&addr,
          sizeof(struct sockaddr))==-1) {
    /* try harder ... */
    true=1;
    if(setsockopt(net->l_fd,SOL_SOCKET,SO_REUSEADDR,&true,sizeof(true))==-1) {
      perror("[network] setsockopt call");
      return N_ERROR;
    }
    dprintf(net->outfd,"[network] reused address\n");
  }

  if(listen(net->l_fd,MAX_LISTEN_QUEUE)==-1) {
    perror("[network] listen call");
    return N_ERROR;
  }

  dprintf(net->outfd,"[network] listen on %s port %d\n",
          inet_ntoa(addr.sin_addr),net->l_port);

  return N_SUCCESS;
}

int network_shutdown(t_net *net) {

  int channel;

  for(channel=0;channel<MAX_CONNECTIONS;channel++)
    if(net->connection[channel].status&C_SOCKET)
      close(net->connection[channel].fd);

  if(close(net->l_fd)==-1) {
    perror("[network] close call");
    return N_ERROR;
  }

  dprintf(net->outfd,"[network] shutdown\n");

  return N_SUCCESS;
}

int network_set_listen_port(t_net *net,in_port_t port) {

  net->l_port=port;

  return N_SUCCESS;
}

int network_manage_connection(t_net *net) {

  int i;
  struct sockaddr_in addr;
  
  for(i=0;i<MAX_CONNECTIONS;i++) {

    if(net->connection[i].status&C_IN_USE) {

      if(net->connection[i].status&C_HANGUP) {
        if(close(net->connection[i].fd)==-1) {
          perror("[network] close call");
          return N_E_CLOSE;
        }
        dprintf(net->outfd,"[network] connection %d closed\n",i);
        net->connection[i].status=0;
      }

      if(net->connection[i].status&C_INFO_A) {

        if(!(net->connection[i].status&C_SOCKET)) {
          if((net->connection[i].fd=socket(AF_INET,SOCK_STREAM,0))==-1) {
            perror("[network] socket call");
            return N_ERROR;
          }
          net->connection[i].status|=C_SOCKET;
        }

        if((!(net->connection[i].status&C_ESTABL))&&
           (net->connection[i].status&C_SOCKET)) {

          memset(&addr,0,sizeof(struct sockaddr));
          addr.sin_family=AF_INET;
          addr.sin_port=htons(net->connection[i].port);
          if(!inet_aton(net->connection[i].ip,&(addr.sin_addr))) {
            perror("[network] inet_aton call");
            return N_ERROR;
          }
          
          if(connect(net->connection[i].fd,(struct sockaddr *)&addr,
                     sizeof(struct sockaddr))==-1) {
            perror("[network] connect call");
            return N_E_CONNECT;
          }

          dprintf(net->outfd,"[network] established connection to ");
          dprintf(net->outfd,"%s port %d on channel %d\n",
                  net->connection[i].ip,net->connection[i].port,i);
          net->connection[i].status|=C_ESTABL;

        }

      }

    }

  }

  return N_SUCCESS;
}

int network_connect(t_net *net,int channel) {

  if(net->connection[channel].status&C_IN_USE) {
    dprintf(net->outfd,"[network] connect failed, channel %02d in use\n",
            channel);
    return N_E_IN_USE;
  }
  if(!(net->connection[channel].status&C_INFO_A)) {
    dprintf(net->outfd,
            "[network] connect failed, missing config for chan %02d\n",channel);
    return N_E_NO_INFO;
  }

  /* try connect & return result */
  net->connection[channel].status|=C_IN_USE;
  return(network_manage_connection(net)); /* could be other channel too */
}

int network_close(t_net *net,int channel) {

  if(!(net->connection[channel].status&C_ESTABL)) {
    dprintf(net->outfd,"[network] close failed, channel %02d not active\n",
            channel);
    return N_E_NC;
  }
  
  net->connection[channel].status|=C_HANGUP;
  return(network_manage_connection(net)); /* could be other channel too */
}

int network_close_all(t_net *net) {

  int channel;

  for(channel=0;channel<MAX_CONNECTIONS;channel++)
    if(net->connection[channel].status&C_ESTABL) network_close(net,channel);

  return N_SUCCESS;
}

int network_set_connection_info(t_net *net,int channel,char *ip,int port) {

  if(net->connection[channel].status&C_IN_USE) {
    dprintf(net->outfd,
            "[network] set connection failed, channel %02d in use\n",channel);
    return N_E_IN_USE;
  }

  strncpy(net->connection[channel].ip,ip,IP_DIGITS);
  net->connection[channel].port=port;
  net->connection[channel].status|=C_INFO_A;

  return N_SUCCESS;
}

int network_select(t_net *net,int channel) {

  int mask;

  if(channel==MAX_CONNECTIONS) mask=0xffffffff;
  else mask=(1<<channel);
  net->sendmask|=mask;

  return N_SUCCESS;
}

int network_deselect(t_net *net,int channel) {

  int mask;

  if(channel==MAX_CONNECTIONS) mask=0;
  else mask=~(1<<channel);
  net->sendmask&=mask;

  return N_SUCCESS;
}

int network_manage_incoming(t_net *net) {

  int channel;
  struct sockaddr_in addr;
  int len;

  for(channel=0;channel<MAX_CONNECTIONS;channel++) {
    if(!(net->connection[channel].status&C_IN_USE)) {
      if((net->connection[channel].fd=accept(net->l_fd,
                                       (struct sockaddr *)&addr,
                                       &len))==-1) {
        perror("[network] accept call");
        return N_E_ACCEPT;
      }
      strncpy(net->connection[channel].ip,inet_ntoa(addr.sin_addr),IP_DIGITS);
      net->connection[channel].port=ntohs(addr.sin_port);
      net->connection[channel].status=C_IN_USE|C_INFO_A|C_SOCKET|C_ESTABL;
      dprintf(net->outfd,
              "[network] established connection from %s:%d, channel %d\n",
              net->connection[channel].ip,net->connection[channel].port,
              channel);
      return channel;
    }
  }

  dprintf(net->outfd,"[network] maximum connections reached\n");
  return N_E_MAXC;
}
 
int network_send(int fd,unsigned char *data,int datasize) {

  int count,left;

  count=0;
  left=datasize;

  while(left) {
    if((count=write(fd,data+datasize-left,left))==-1) {
      perror("[network] write call");
      return N_ERROR;
    }
    left-=count;
  }

  return N_SUCCESS;
}

int network_receive(int fd,unsigned char *data,int datasize) {

  int count,left;

  count=0;
  left=datasize;

  while(left) {
    if((count=read(fd,data,datasize))==-1) {
      perror("[network] read call");
      return N_ERROR;
    }
    left-=count;
  }

  return datasize;
}

int network_receive_chan(t_net *net,int chan,unsigned char *data,int datasize) {

  return(network_receive(net->connection[chan].fd,data,datasize));
}

int network_udp_listen_init(t_net *net) {

  struct sockaddr_in addr;
  int true;

  if((net->l_udp_fd=socket(AF_INET,SOCK_DGRAM,0))==-1) {
    perror("[network] socket call (udp-receive)");
    return N_ERROR;
  }

  memset(&addr,0,sizeof(struct sockaddr));
  addr.sin_family=AF_INET;
  addr.sin_port=htons(net->l_udp_port);
  addr.sin_addr.s_addr=INADDR_ANY;
  if(bind(net->l_udp_fd,(struct sockaddr *)&addr,sizeof(struct sockaddr))==-1) {
    /* try harder ... */
    true=1;
    if(setsockopt(net->l_udp_fd,SOL_SOCKET,SO_REUSEADDR,
                  &true,sizeof(true))==-1) {
      perror("[network] setsockopt call (udp)");
      return N_ERROR;
    }
    dprintf(net->outfd,"[network] reused address (udp)\n");
  }

  dprintf(net->outfd,"[network] listening on port %d (udp)\n",
          net->l_udp_port);

  if((net->s_udp_fd=socket(AF_INET,SOCK_DGRAM,0))==-1) {
    perror("[network] socket call (udp-send)");
    return N_ERROR;
  }

  return N_SUCCESS;
}

int network_udp_receive(t_net *net,int channel, unsigned char *data,int count) {

  struct sockaddr_in addr;
  socklen_t len;

  if((count=recvfrom(net->l_udp_fd,data,count,0,
                     (struct sockaddr *)&addr,&len))==-1) {
    perror("[network] recvfrom call");
    return N_ERROR;
  }

  if(strncmp(net->connection[channel].ip,inet_ntoa(addr.sin_addr),IP_DIGITS)) {
    dprintf(net->outfd,"[network] packet from unknown: %s\n",
            inet_ntoa(addr.sin_addr));
    return N_UDP_WRONG_SENDER;
  }

  return N_SUCCESS;
}

int network_udp_send(t_net *net,int channel, unsigned char *data,int size) {

  int count,left;
  struct sockaddr_in addr;

  count=0;
  left=count;

  memset(&addr,0,sizeof(struct sockaddr));
  addr.sin_family=AF_INET;
  addr.sin_port=htons(net->l_udp_port);
  inet_aton(net->connection[channel].ip,&(addr.sin_addr));

  while(left) {
    if((count=sendto(net->s_udp_fd,data+size-left,left,0,
                     (struct sockaddr *)&addr,sizeof(struct sockaddr)))==-1) {
      perror("[network] sendto call");
      return N_ERROR;
    }
    left-=count;
  }

  return N_SUCCESS;
}

int network_udp_shutdown(t_net *net) {
  
  if(close(net->l_udp_fd)==-1) {
    perror("[network] close call (udp-receive)");
    return N_ERROR;
  }

  if(close(net->s_udp_fd)==-1) {
    perror("[network] close call (udp-send)");
    return N_ERROR;
  }

  return N_SUCCESS;
}
